"""Serialization utility operations"""

import ast
import json
import logging
from enum import Enum
from typing import Any, List, Optional, Set

import yaml

logger = logging.getLogger(__name__)


class LongStringDumper(yaml.SafeDumper):
    """Custom YAML dumper to handle long strings."""

    def represent_scalar(
        self,
        tag: str,
        value: str,
        style: Optional[str] = None,
    ) -> yaml.nodes.ScalarNode:
        """Represent scalar values in YAML."""
        # Replace any escape sequences to format the output as desired
        if (
            len(value) > 150 or "\n" in value
        ):  # Adjust the threshold for long strings as needed
            style = "|"
            # converts all newline escapes to actual representations
            value = "\n".join([line.rstrip() for line in value.split("\n")])
        else:
            # converts all newline escapes to actual representations
            value = "\n".join([line.rstrip() for line in value.split("\n")]).rstrip()

        return super().represent_scalar(tag, value, style)


def make_serializable(obj: Any) -> Any:
    """Recursively convert non-serializable objects in a dict/list to strings."""
    if isinstance(obj, dict):
        return {
            make_serializable(key): make_serializable(value)
            for key, value in obj.items()
        }
    elif isinstance(obj, list):
        return [make_serializable(item) for item in obj]
    elif isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    else:
        # For sets, tuples, custom types, etc.
        return str(obj)


def export_to_dict(
    data: object | dict,
    ignore_keys: Optional[List[str]] = None,
) -> dict:
    """Export an object to a dictionary, ignoring specified keys and handling cycles.

    Args:
        data: The object or dictionary to serialize.
        ignore_keys: Keys to exclude from serialization (default: ["__jac__"])

    Returns:
        A dictionary representation of the input.
    """
    if ignore_keys is None:
        ignore_keys = ["__jac__"]

    memo: Set[int] = set()  # Track object IDs for cycle detection

    def _convert(obj: object) -> object:
        # Handle cycles
        obj_id = id(obj)
        if obj_id in memo:
            return "<cycle detected>"
        memo.add(obj_id)
        try:
            # 1. Basic immutable types
            if obj is None or isinstance(obj, (bool, int, float, str)):
                return obj

            # 2. Handle enums by their value
            if Enum is not None and isinstance(obj, Enum):
                return _convert(obj.value)  # Serialize enum value

            # 3. Handle namedtuples
            if hasattr(obj, "_asdict") and callable(obj._asdict):
                return _convert(obj._asdict())

            # 4. Dictionaries: apply ignore_keys and recurse
            if isinstance(obj, dict):
                return {k: _convert(v) for k, v in obj.items() if k not in ignore_keys}

            # 5. Lists, tuples, sets: convert to list and recurse
            if isinstance(obj, (list, tuple, set, frozenset)):
                return [_convert(item) for item in obj]

            # 6. Generic objects with __dict__
            if hasattr(obj, "__dict__"):
                return _convert(obj.__dict__)

            # 7. Fallback: string representation
            return str(obj)

        finally:
            memo.discard(obj_id)  # Clean up after processing

    result = _convert(data)
    # Ensure top-level output is a dictionary
    return result if isinstance(result, dict) else {"value": result}


def safe_json_dump(data: dict) -> Optional[str]:
    """Safely convert a dictionary with mixed types to a JSON string for logs."""
    if not isinstance(data, dict):
        logger.error("Input to safe_json_dump must be a dictionary.")
        return None

    def serialize(obj: dict) -> dict:
        """Recursively convert strings within complex objects."""

        def wrap_content(value: object) -> object:
            # Return value wrapped in a dictionary with key 'content' if it's a str, int or float
            return {"content": value} if isinstance(value, (str, int, float)) else value

        def process_dict(d: dict) -> dict:
            for key, value in d.items():
                if isinstance(value, dict):
                    d[key] = process_dict(value)
                elif isinstance(value, list):
                    # If the list contains dictionaries, recursively process each
                    d[key] = [
                        process_dict(item) if isinstance(item, dict) else item
                        for item in value
                    ]
                else:
                    d[key] = wrap_content(value)
            return d

        # Create a deep copy of the original dict to avoid mutation
        import copy

        result = copy.deepcopy(obj)

        return process_dict(result)

    try:
        # Attempt to serialize the dictionary
        return json.dumps(serialize(data))
    except (TypeError, ValueError) as e:
        # Handle serialization errors
        logger.error(f"Serialization error: {str(e)}")
        return None


def convert_str_to_json(text: str) -> dict | None:
    """Convert a string to a JSON object."""
    if isinstance(text, str):
        text = text.replace("```json", "")
        text = text.replace("```", "")
    try:
        if isinstance(text, (dict, list)):
            return text
        else:
            return json.loads(text)
    except (json.JSONDecodeError, TypeError):
        try:
            return ast.literal_eval(text)
        except (SyntaxError, ValueError) as e:
            if "'{' was never closed" in str(e):
                text = text + "}"
                return json.loads(text)
            else:
                logger.error(e)
                return None


def yaml_dumps(data: Optional[dict]) -> Optional[str]:
    """Converts and formats nested dict to YAML string, handling PyYAML errors."""
    if not data:
        return None

    try:
        safe_data = make_serializable(data)
        yaml_output = yaml.dump(
            safe_data,
            Dumper=LongStringDumper,
            allow_unicode=True,
            default_flow_style=False,
            sort_keys=False,
        )
        if not yaml_output or yaml_output.strip() in ("", "---"):
            return None
        return yaml_output

    except Exception as e:
        logger.error(f"Error dumping YAML: {e}")
        return None



"""Utils to chunk text"""

import logging
import re
from typing import List

logger = logging.getLogger(__name__)


def chunk_long_message(
    message: str, max_length: int = 1024, chunk_length: int = 1024
) -> List[str]:
    """
    Splits a long message into smaller chunks of no more than chunk_length characters,
    ensuring no single chunk exceeds max_length.

    Args:
        message: The text to chunk
        max_length: Maximum allowed length for any chunk
        chunk_length: Target length for chunks

    Returns:
        List of message chunks
    """
    if len(message) <= max_length:
        return [message]

    # Initialize variables
    final_chunks = []
    current_chunk = ""
    current_chunk_length = 0

    # Split the message into words while preserving newline characters
    words = re.findall(r"\S+\n*|\n+", message)
    words = [word for word in words if word.strip()]  # Filter out empty strings

    for word in words:
        word_length = len(word)

        if current_chunk_length + word_length + 1 <= chunk_length:
            # Add the word to the current chunk
            if current_chunk:
                current_chunk += " "
            current_chunk += word
            current_chunk_length += word_length + 1
        else:
            # If the current chunk is full, add it to the list of chunks
            final_chunks.append(current_chunk)
            current_chunk = word  # Start a new chunk with the current word
            current_chunk_length = word_length

    if current_chunk:
        # Add the last chunk if it's non-empty
        final_chunks.append(current_chunk)

    return final_chunks



"""
jivas package initialization.

JIVAS is an Agentic Framework for rapidly prototyping and deploying graph-based, AI solutions.
"""

from importlib.metadata import version

__version__ = version("jivas")



"""Commit utilities for Jac memory."""

from jac_cloud.core.archetype import BaseAnchor, BulkWrite
from jac_cloud.jaseci.datasources import Collection
from jac_cloud.plugin.jaseci import JacPlugin as Jac
from jaclang.runtimelib.constructs import Archetype


def commit(anchor: BaseAnchor | None = None) -> None:
    """Commit all data from memory to datasource."""
    if anchor:
        if isinstance(anchor, Archetype):
            anchor = anchor.__jac__
        anchor.build_query(bulk_write := BulkWrite())
    else:
        bulk_write = Jac.get_context().mem.get_bulk_write()

    if bulk_write.has_operations:
        if session := Jac.get_context().mem.__session__:
            bulk_write.execute(session)
        else:
            with Collection.get_session() as session, session.start_transaction():
                bulk_write.execute(session)



"""Module for Jivas Embeddings."""

import json
from typing import List

from langchain_core.embeddings import Embeddings
from openai import OpenAI
from transformers import AutoTokenizer


class JivasEmbeddings(Embeddings):
    """Class for handling Jivas Embeddings."""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        model: str = "intfloat/multilingual-e5-large-instruct",
    ) -> None:
        """Initialize the JivasEmbeddings class."""
        # init args
        self.model = model
        self.model_name = "-".join(model.split("/"))
        self.base_url = base_url
        self.api_key = api_key

        # create client
        self.client = OpenAI(api_key=api_key, base_url=base_url)

        # Load the tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.token_limit = self.tokenizer.model_max_length

    def trim_text_if_needed(self, text: str) -> str:
        """Trim text if it exceeds the token limit."""
        tokenized = self.tokenizer(text, return_tensors="pt")
        num_tokens = len(tokenized["input_ids"][0])

        if num_tokens > self.token_limit:
            truncated_tokens = tokenized["input_ids"][0][: self.token_limit]
            return self.tokenizer.decode(truncated_tokens, skip_special_tokens=True)

        return text

    def embed_documents(
        self, texts: List[str], handle_overflow: bool = False
    ) -> List[List[float]]:
        """Embed search documents."""
        if handle_overflow:
            texts = [self.trim_text_if_needed(text) for text in texts]

        try:
            # grab embeddings
            response = self.client.embeddings.create(input=texts, model=self.model_name)
            # set response to json
            response = json.loads(response.json())
            # return embeddings
            return [embd["embedding"] for embd in response["data"]]
        except Exception as e:
            # Handle potential errors
            print(f"Error embedding documents: {e}")
            return []

    def embed_query(self, text: str, handle_overflow: bool = False) -> List[float]:
        """Embed query text."""
        if handle_overflow:
            text = self.trim_text_if_needed(text)

        try:
            # grab embeddings
            response = self.client.embeddings.create(input=text, model=self.model_name)
            # set response to json
            response = json.loads(response.json())
            # return embeddings
            return response["data"][0]["embedding"]
        except Exception as e:
            # Handle potential errors
            print(f"Error embedding query: {e}")
            return []



"""Operations to clean up actions"""

import logging
import os
import shutil
from typing import Dict, List

from ..text.formatting import normalize_text

logger = logging.getLogger(__name__)


def clean_action(namespace_package_name: str) -> bool:
    """Completely removes a specific action folder.

    Args:
        namespace_package_name: The namespace and package name in format 'namespace_folder/action_folder'

    Returns:
        bool: True if the action folder was successfully removed, False otherwise
    """
    actions_root_path = os.environ.get("JIVAS_ACTIONS_ROOT_PATH", "actions")
    if os.path.isdir(actions_root_path):
        logger.info(f"Running clean on action {namespace_package_name}")

        # Construct full path to the action folder
        action_path = os.path.join(actions_root_path, namespace_package_name)

        try:
            if os.path.exists(action_path):
                # Remove the entire action folder and its contents
                shutil.rmtree(action_path)
                logger.info(f"Successfully removed action folder: {action_path}")
                return True
            else:
                logger.warning(f"Action folder not found: {action_path}")
                return False
        except Exception as e:
            logger.error(f"Failed to remove action folder {action_path}: {str(e)}")
            return False
    else:
        logger.error(f"Actions root directory not found: {actions_root_path}")
        return False


def clean_context(
    node_context: Dict, archetype_context: Dict, ignore_keys: List[str]
) -> Dict:
    """
    Cleans and sanitizes node_context by:
    - Removing keys that match in node_context and archetype_context
    - Removing empty values except boolean `False`
    - Removing keys listed in ignore_keys.

    Args:
        node_context: Existing snapshot of spawned node
        archetype_context: Original archetype context variables
        ignore_keys: List of keys to remove

    Returns:
        Sanitized dictionary
    """
    # Check for matching keys and remove them if they match
    for key in list(archetype_context.keys()):
        if key in node_context:
            if isinstance(node_context[key], str):
                str1 = normalize_text(node_context[key])
                str2 = normalize_text(archetype_context[key])

                if str1 == str2:
                    del node_context[key]
            else:
                if node_context[key] == archetype_context[key]:
                    del node_context[key]

    # Prepare keys to remove
    to_remove_key = list(ignore_keys)

    # Remove empty values (except boolean False)
    for key in list(node_context.keys()):
        if not node_context[key] and not isinstance(node_context[key], bool):
            to_remove_key.append(key)

    # Remove specified keys
    for key in to_remove_key:
        node_context.pop(key, None)

    return node_context



"""node_get paginating node collections in Jivas."""

from jac_cloud.core.archetype import BaseCollection, NodeAnchor
from jac_cloud.plugin.jaseci import JacPlugin as Jac


def node_get(query_filter: dict | None = None) -> list:
    """Retrieve a list of nodes from the 'node' collection based on the query filter."""

    if query_filter is None:
        return []

    # Execute the query
    node_refs = [
        NodeAnchor.ref(f"n::{str(nd.get("_id"))}")
        for nd in BaseCollection.get_collection("node").find(query_filter)
    ]

    if node_refs:
        nodes = [n.archetype for n in Jac.get_context().mem.find(node_refs)]
        return nodes

    return []



"""NodePager class for paginating node collections in Jivas."""

from jac_cloud.jaseci.datasources.collection import Collection

"""
# Initialize pager
pager = Pager(NodeAnchor.Collection, page_size=10, current_page=2)

# Get a page of results
doc_entries = pager.get_page({"name": "DocFileEntry"})

# get all info as a dict
pagination_info = pager.to_dict()

"""


class NodePager:
    """A class to handle pagination of nodes in a collection."""

    def __init__(
        self,
        collection: Collection,
        page_size: int = 10,
        current_page: int = 1,
    ) -> None:
        """Initialize the NodePager with a collection, optional root, page size, and current page."""
        self.collection = collection
        self.page_size = page_size
        self.current_page = current_page
        self.total_items = 0
        self.total_pages = 1
        self.has_previous = False
        self.has_next = False

    def get_page(self, query_filter: dict | None = None) -> list:
        """Retrieve a paginated list of nodes based on the query filter."""

        if query_filter is None:
            query_filter = {}

        # Get total count of items matching the filter
        self.total_items = self.collection.count(query_filter)

        # Calculate total pages
        self.total_pages = max(
            1, (self.total_items + self.page_size - 1) // self.page_size
        )

        # Ensure current_page is within bounds
        self.current_page = max(1, min(self.current_page, self.total_pages))

        # Calculate pagination flags
        self.has_previous = self.current_page > 1
        self.has_next = self.current_page < self.total_pages

        # Calculate skip value
        skip = (self.current_page - 1) * self.page_size

        # Execute the query
        items = self.collection.find(
            query_filter,
            skip=skip,
            limit=self.page_size,
        )

        if items:
            nodes = [n.archetype for n in items]
            return nodes

        return []

    @property
    def previous_page(self) -> int | None:
        """Return the previous page number if available, otherwise None."""
        return self.current_page - 1 if self.has_previous else None

    @property
    def next_page(self) -> int | None:
        """Return the next page number if available, otherwise None."""
        return self.current_page + 1 if self.has_next else None

    def to_dict(self) -> dict:
        """Return a dictionary representation of the pagination state."""
        return {
            "total_items": self.total_items,
            "total_pages": self.total_pages,
            "current_page": self.current_page,
            "has_previous": self.has_previous,
            "has_next": self.has_next,
            "page_size": self.page_size,
        }



"""Utils to handle parsing"""


def extract_first_name(full_name: str) -> str:
    """
    Extract the first name from a full name.

    Args:
        full_name: The complete name string

    Returns:
        The extracted first name
    """
    # List of common titles to be removed
    titles = ["Mr.", "Mrs.", "Ms.", "Miss", "Dr.", "Prof.", "Sir", "Madam", "Mx."]

    # Split the full name into parts
    name_parts = full_name.split()

    # Remove any titles from the list of name parts
    name_parts = [part for part in name_parts if part not in titles]

    # Return the first element as the first name if it's available
    return name_parts[0] if name_parts else ""



"""Setup script for jivas"""

import os

from setuptools import find_packages, setup


def get_version() -> str:
    """Get the package version from the __init__ file."""
    version_file = os.path.normpath(
        os.path.join(os.path.dirname(__file__), "..", "__init__.py")
    )
    with open(version_file) as f:
        for line in f:
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Version not found.")


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="jivas",
    version=get_version(),
    description="JIVAS is an Agentic Framework for rapidly prototyping and deploying graph-based, AI solutions.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="TrueSelph Inc.",
    author_email="admin@trueselph.com",
    url="https://github.com/TrueSelph/jivas",
    keywords=["jivas"],
    packages=find_packages(
        include=[
            "jivas",
            "jivas.*",
        ]
    ),
    include_package_data=True,
    package_data={"jivas": []},
    python_requires=">=3.12.0",
    install_requires=[
        "jvcli<=2.1.20",
        "jvclient<=2.1.20",
        "jvserve<=2.1.20",
        "jac-cloud==0.2.7",
        "jaclang==0.8.7",
        "pytz==2025.2",
        "types-pytz==2025.2.0.20250809",
        "schedule==1.2.2",
        "openai==1.109.1",
        "langchain==0.3.27",
        "langchain-community==0.3.27",
        "langchain-core==0.3.79",
        "langchain-experimental==0.3.4",
        "langchain-openai==0.3.19",
        "langchain-text-splitters==0.3.11",
        "langsmith==0.4.38",
        "setuptools==80.9.0",
        "transformers==4.57.1",
        "ftfy==6.3.1",
    ],
    extras_require={
        "dev": [
            "pre-commit",
            "pytest",
            "pytest-mock",
            "pytest-cov",
            "coverage",
        ]
    },
)
